// -*-c++-*-

/***************************************************************************
                                creator.h 
                             -------------------
                     Classes used by factory to create objects
    begin                : 2002-10-08
    copyright            : (C) 2002 by The RoboCup Soccer Simulator 
                           Maintenance Group.
    email                : sserver-admin@lists.sourceforge.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU LGPL as published by the Free Software  *
 *   Foundation; either version 2 of the License, or (at your option) any  *
 *   later version.                                                        *
 *                                                                         *
 ***************************************************************************/


#ifndef RCSS_CREATOR_H
#define RCSS_CREATOR_H

#include <boost/shared_ptr.hpp>
#include <boost/weak_ptr.hpp>
#include "depobj.h"
#include "libloader.h"

namespace rcss
{
    // CreateNew

    template< class RV, class P = void >
    class
    Create
    {
    public:
        typedef Create< RV, P > Base;
        typedef P Params;
        typedef RV RVal;
        
        virtual
        ~Create()
        {}

        virtual
        RVal
        operator()( P params ) = 0;
    };

    template< class RV >
    class
    Create< RV, void >
    {
    public:
        typedef Create< RV, void > Base;
        typedef void Params;
        typedef RV RVal;
        
        virtual
        ~Create()
        {}

        virtual
        RVal
        operator()() = 0;
    };

    template< class PO, class RV, class P = void >
    class
    CreateNew
        : public Create< RV, P >
    {
    public:
        typedef PO PolyObj;
        typedef RV RVal;

        RVal
        operator()( P params )
        { return RVal( new PolyObj( params ) ); }
    };

    template< class PO, class RV >
    class
    CreateNew< PO, RV, void >
        : public Create< RV >
    {
    public:
        typedef PO PolyObj;
        typedef RV RVal;

        RVal
        operator()()
        { return RVal( new PolyObj() ); }
    };

 
    template< class PO, class OBJ, class P >
    class 
    CreateNew< PO, rcss::DepObj< OBJ, rcss::Lib >, P >
        : public Create< rcss::DepObj< OBJ, rcss::Lib >, P >
    {
    public:
        typedef PO PolyObj;
        typedef rcss::DepObj< OBJ, rcss::Lib > RVal;

        CreateNew( const rcss::WeakLib& dep )
            : m_dep( dep )
        {}

        RVal
        operator()( P params )
        { return RVal( OBJ( new PolyObj( params ) ), m_dep ); }
        
    private:
        rcss::WeakLib m_dep;
    };


    template< class PO, class OBJ >
    class 
    CreateNew< PO, rcss::DepObj< OBJ, rcss::Lib >, void >
        : public Create< rcss::DepObj< OBJ, rcss::Lib >, void >
    {
    public:
        typedef PO PolyObj;
        typedef rcss::DepObj< OBJ, rcss::Lib > RVal;

        CreateNew( const rcss::WeakLib& dep )
            : m_dep( dep )
        {}

        RVal
        operator()()
        { return RVal( OBJ( new PolyObj( params ) ), m_dep ); }
        
    private:
        rcss::WeakLib m_dep;
    };



    // CreateSing

    template< class PO, class RV, class P = void >
    class
    CreateSing
        : public Create< RV, P >
    {
    public:
        typedef PO PolyObj;
        typedef RV RVal;

        RVal
        operator()( P params )
        { return PolyObj::instance( params ); }
    };

    template< class PO, class RV >
    class
    CreateSing< PO, RV, void >
        : public Create< RV >
    {
    public:
        typedef PO PolyObj;
        typedef RV RVal;

        RVal
        operator()()
        { return PolyObj::instance(); }
    };


   template< class PO, class BO, class P >
    class 
    CreateSing< PO, rcss::DepObj< BO, rcss::Lib >, P >
        : public Create< rcss::DepObj< BO, rcss::Lib >, P >
    {
    public:
        typedef PO PolyObj;
        typedef rcss::DepObj< BO, rcss::Lib > RVal;

        CreateSing( const rcss::WeakLib& dep )
            : m_dep( dep )
        {}

        RVal
        operator()( P params )
        {
            return RVal( PolyObj::instance( params ), m_dep );
        }
        
    private:
        rcss::WeakLib m_dep;
    };
    
    template< class PO, class BO >
    class 
    CreateSing< PO, rcss::DepObj< BO, rcss::Lib >, void >
        : public Create< rcss::DepObj< BO, rcss::Lib >, void >
    {
    public:
        typedef PO PolyObj;
        typedef rcss::DepObj< BO, rcss::Lib > RVal;

        CreateSing( const rcss::WeakLib& dep )
            : m_dep( dep )
        {}

        RVal
        operator()()
        {
            return RVal( PolyObj::instance(), m_dep );
        }
        
    private:
        rcss::WeakLib m_dep;
    };
}

#endif
