// -*-c++-*-

/***************************************************************************
                                factory.h 
                             -------------------
    Template singleton for creating polymorphic objects based on some idx
    begin                : 2002-10-08
    copyright            : (C) 2002 by The RoboCup Soccer Simulator 
                           Maintenance Group.
    email                : sserver-admin@lists.sourceforge.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU LGPL as published by the Free Software  *
 *   Foundation; either version 2 of the License, or (at your option) any  *
 *   later version.                                                        *
 *                                                                         *
 ***************************************************************************/


#ifndef RCSSFACTORY_H
#define RCSSFACTORY_H

#include <memory>
#include <map>
#include <deque>
#include <list>
#include <exception>
#include <iostream>
#include <boost/shared_ptr.hpp>
#include <typeinfo>

namespace rcss
{

    
/*!
//===================================================================
//
//  CLASS: ClassNotFound
//
//  DESC: Get's throw in Factory when the class requested cannot be
//  found.
//
//===================================================================
*/
 
    template< class I = const char* >
    class ClassNotFound
        : public std::exception
    {
    public:
        typedef I Index;

        ClassNotFound( Index idx ) throw()
            : m_idx( idx )
        {}

        ~ClassNotFound() throw()
        {}

        const char*
        what() const throw()
        { return "This class is not registered with this type of creator"; }

        Index
        index() const throw()
        { return m_idx; }
    private:
        Index m_idx;
    };

    /*!
      //===========================================================
      //
      //  CLASS: RegHolderImpl
      //
      //  DESC: Base type for AutoReger
      //
      //===========================================================
    */

    class RegHolderImpl
    {
    public:
        RegHolderImpl()
        {}
        
        virtual 
        ~RegHolderImpl()
        {}
    };
    
    typedef std::auto_ptr< RegHolderImpl > RegHolder;


    /*!
      //===========================================================
      //
      //  CLASS: AutoReger
      //
      //  DESC: Used for automatic registration.
      //
      //===========================================================
    */

    template< class OF >
    class AutoReger
        : public RegHolderImpl
    {
    public:
        typedef typename OF::Creator Creator;
        typedef typename OF::Index Index;
        
        AutoReger( Creator creator, const Index& idx )
            : m_idx( idx )
        { OF fact; fact.reg( creator, idx ); }
        
        virtual
        ~AutoReger()
        { OF fact; fact.dereg( m_idx ); }
    private:
        template< class OF2 >
        AutoReger( const AutoReger< OF2 >& ); // not used

        template< class OF2 >
        AutoReger&
        operator=( const AutoReger< OF2 >& ); // not used

    private:
        const Index& m_idx;
    };

/*!
//===================================================================
//
//  CLASS: ObjFact
//
//  DESC: An Generic Object Factory (aka Class Store)
//
//===================================================================
*/
 
    template< class Cre, class I = std::string, class Com = std::less< I > >
    class ObjFact
    {
    public:
        typedef Cre Creator;
        typedef I Index;
        typedef Com Compare;
              
    private:
        class ObjFactImpl
        {
        private:        
            typedef std::map< Index, Creator, Compare > Map;
            
        public:       
            ObjFactImpl()
            {}

            ~ObjFactImpl()
            {}
            
            const std::type_info&
            type() const
            { return typeid( *this ); }
 
            bool
            reg( Creator c, const Index& i )
            { 
                if( m_creators.find( i ) != m_creators.end() ) 
                    return false;
                m_creators.insert( std::pair< Index, Creator >( i, c ) ); 
                return true;
            }

            void
            dereg( const Index& i )
            {
                m_creators.erase( i );
            }
            
            Creator
            getCreator( const Index& i )
            {
                typename Map::iterator iter = m_creators.find( i );
                if( iter != m_creators.end() )
                    return iter->second;
                throw ClassNotFound< Index >( i );
            }
            
            std::list< Index >
            list()
            {
                std::list< Index > rval;
                for( typename Map::iterator i = m_creators.begin(); i != m_creators.end(); ++i )
                    rval.push_back( i->first );
                return rval;
            }
            
            std::ostream&
            printList( std::ostream& o = std::cout)
            {
                for( typename Map::iterator i = m_creators.begin(); i != m_creators.end(); ++i )
                    o << "\t" << i->first << std::endl;
                return o;
            }

            static 
            ObjFactImpl&
            instance()
            { static ObjFactImpl rval; return rval; }

            unsigned int
            mapSize() const
            { return m_creators.size(); }
        private:
            Map m_creators;
        };

    public:
        ObjFact()
        {}

        ~ObjFact()
        {}

        bool
        reg( Creator c, const Index& i )
        {
            if( !s_impl )
                s_impl = new ObjFactImpl();
            return s_impl->reg( c, i ); 
        }

        void
        dereg( const Index& i )
        { 
            if( s_impl )
            {
                s_impl->dereg( i );
                if( s_impl->mapSize() == 0 )
                { delete s_impl; s_impl = NULL; }
            }
        }
            
        Creator
        getCreator( const Index& i )
        { 
            if( s_impl )
                return s_impl->getCreator( i );
            throw ClassNotFound< Index >( i );
        }
        
        std::list< Index >
        list()
        {
            if( s_impl )
                return s_impl->list(); 
            return std::list< Index >();
        }
            
        std::ostream&
        printList( std::ostream& o = std::cout )
        {
            o << "listing classes (";
            if( s_impl )
            {
                o << s_impl->mapSize() << "):\n";
                s_impl->printList( o ); 
            }
            else
				o << "0):\n";
			return o << "done\n";
        }

        static
        RegHolder
        autoReg( Creator c, const Index& i )
        { return RegHolder( new AutoReger< ObjFact< Creator, Index, Compare > >( c, i ) ); }

    private:
        static ObjFactImpl* s_impl;
    };

    template< class Cre, class I, class Com >
    typename ObjFact< Cre, I, Com >::ObjFactImpl* ObjFact< Cre, I, Com >::s_impl = NULL;


}


#endif
