// -*-c++-*-

/***************************************************************************
                                 gzfstream.h
                      Compression and decompression file streams
                             -------------------
    begin                : 31-OCT-2001
    copyright            : (C) 2001, 2002 by The RoboCup Soccer Server 
                           Maintenance Group.
    email                : sserver-admin@lists.sourceforge.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU LGPL as published by the Free Software  *
 *   Foundation; either version 2 of the License, or (at your option) any  *
 *   later version.                                                        *
 *                                                                         *
 ***************************************************************************/

#ifndef GZFSTREAM_H
#define GZFSTREAM_H

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif


#ifdef HAVE_LIBZ
#include <zlib.h>
#include <iostream>

#ifdef HAVE_SSTREAM
#include <sstream>
#else
#include <strstream>
#endif

#ifndef EOF
#define EOF (-1)
#endif

namespace rcss
{
  namespace gz
  {
    class gzfstreambuf
      : public std::streambuf
    {
    public:
      typedef int int_type;
      typedef char char_type;

      gzfstreambuf ( int bufsize = 512 )
        : M_open ( false ),
          M_gzfile ( NULL ),
          bufsize_(bufsize),
          inbuf_(NULL),
          outbuf_(NULL),
          remained_(0)
      {
      }

      ~gzfstreambuf ()
      {
        if ( is_open () )
          close ();

        delete [] inbuf_;
        delete [] outbuf_;
      }

      bool is_open ()
      { 
        return M_open;
      }  

      bool
      open( const char* path, 
            int level = Z_DEFAULT_COMPRESSION,
            int strategy = Z_DEFAULT_STRATEGY )
      {
#ifdef HAVE_SSTREAM
        std::ostringstream mode_str;
#else
        std::ostrstream mode_str;
#endif

        mode_str << "wb" << level;
    
        if ( strategy == Z_FILTERED )
          mode_str << "f";
        else if ( strategy == Z_HUFFMAN_ONLY )
          mode_str << "h";

#ifndef HAVE_SSTREAM    
        mode_str << std::ends;
#endif

#ifdef HAVE_SSTREAM    
        M_gzfile = gzopen ( path, mode_str.str().c_str() );
#else
        M_gzfile = gzopen ( path, mode_str.str() );
        mode_str.freeze( false );
#endif
        M_open = ( M_gzfile != NULL );
        return true;
      }

      void open ( int fd,
                  int level = Z_DEFAULT_COMPRESSION,
                  int strategy = Z_DEFAULT_STRATEGY )
      {
#ifdef HAVE_SSTREAM
        std::ostringstream mode_str;
#else
        std::ostrstream mode_str;
#endif
    
        mode_str << "wb" << level;
    
        if ( strategy == Z_FILTERED )
          mode_str << "f";
        else if ( strategy == Z_HUFFMAN_ONLY )
          mode_str << "h";
    
#ifndef HAVE_SSTREAM    
        mode_str << std::ends;
#endif
    
#ifdef HAVE_SSTREAM    
        M_gzfile = gzdopen ( fd, mode_str.str().c_str() );
#else
        M_gzfile = gzdopen ( fd, mode_str.str() );
        mode_str.freeze( false );
#endif
        M_open = ( M_gzfile != NULL );
      }

      void close ()
      {
        _flush ();
        gzclose ( M_gzfile );
        M_gzfile = NULL;
        M_open = false;
      }

    protected:
      void _flush()
      {
        int size = (pptr () - outbuf_) * sizeof(char_type);
        if ( size > 0 )
          gzwrite ( M_gzfile, outbuf_, size );
      }

      int_type overflow(int_type c = EOF)
      {
        // this method is supposed to flush the put area of the buffer
        // to the I/O device

        // if the buffer was not already allocated nor set by user,
        // do it just now
        if (pptr() == NULL)
          {
            outbuf_ = new char_type[bufsize_];
          }
        else
          {
            _flush();
          }
        setp(outbuf_, outbuf_ + bufsize_);
        if (c != EOF)
          sputc(c);
        return 0;
      }

      int sync()
      {
        // just flush the put area
        if (pptr() != NULL)
          {
            _flush();
            setp(outbuf_, outbuf_ + bufsize_);
            return 0;
          }
        return 0;
      }

      int_type underflow()
      {
        // this method is supposed to read some bytes from the I/O device

        // if the buffer was not already allocated nor set by user,
        // do it just now
        if (gptr() == NULL)
          {
            inbuf_ = new char_type[bufsize_];
          }

        if (remained_ != 0)
          inbuf_[0] = remainedchar_;

        int readn = gzread ( M_gzfile, (void*)(inbuf_ + remained_),
                             bufsize_ * sizeof(char_type) - remained_);

        // if (readn == 0 && remained_ != 0)
        // error - there is not enough bytes for completing
        // the last character before the end of the stream
        // - this can mean error on the remote end

        if (readn == 0)
          return EOF;

        int totalbytes = readn + remained_;
        setg(inbuf_, inbuf_,
             inbuf_ + totalbytes / sizeof(char_type));

        remained_ = totalbytes % sizeof(char_type);
        if (remained_ != 0)
          remainedchar_ = inbuf_[totalbytes / sizeof(char_type)];

        return sgetc();
      }

    private:

      // not for use
      gzfstreambuf(const gzfstreambuf&)
        : std::streambuf()
      {}

      gzfstreambuf& operator=(const gzfstreambuf&)
      { return *this; }

      bool M_open;
      gzFile M_gzfile;
      std::streamsize bufsize_;
      char_type *inbuf_;
      char_type *outbuf_;
      int remained_;
      char_type remainedchar_;
    };

    class gzfstream :
      public gzfstreambuf,
      public std::iostream
    {
    public:

      gzfstream( int buffer_size = 512 )
        : gzfstreambuf( buffer_size ),
          std::iostream( this )
      {
      }

    private:
      // not for use
      gzfstream(const gzfstream&);
      gzfstream& operator=(const gzfstream&);
    };
  } // namespace gz
} // namespace rcss
#endif //HAVE_LIBZ

#endif 
