// -*-c++-*-

/***************************************************************************
                                 gzstream.h
                      Compression and decompression streams
                             -------------------
    begin                : 11-JUL-2002
    copyright            : (C) 2002 by The RoboCup Soccer Server 
                           Maintenance Group.
    email                : sserver-admin@lists.sourceforge.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU LGPL as published by the Free Software  *
 *   Foundation; either version 2 of the License, or (at your option) any  *
 *   later version.                                                        *
 *                                                                         *
 ***************************************************************************/

#ifndef GZSTREAM_H
#define GZSTREAM_H

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef HAVE_LIBZ
#include <zlib.h>
#endif

#include <iostream>

#ifndef EOF
#define EOF (-1)
#endif

namespace rcss
{
  namespace gz
  {
    class gzstreambuf
        : public std::streambuf
    {
    public:
      typedef int int_type;
      typedef char char_type;

      gzstreambuf( std::streambuf& strm,
                   unsigned int bufsize = 8192 )
        : std::streambuf(),
          M_strmbuf( strm ),
          M_output_stream( NULL ),
          M_input_stream( NULL ),
          M_buffer_size( bufsize ),
          M_read_buffer( NULL ),
          M_input_buffer( NULL ),
          M_output_buffer( NULL ),
          M_write_buffer( NULL ),
          M_remained( 0 ),
#ifdef HAVE_LIBZ
          M_compression_stream( NULL ),
          M_decompression_stream( NULL ),
#endif
          M_level(  
#ifdef HAVE_LIBZ
                  Z_DEFAULT_COMPRESSION
#else
                  -1
#endif
                  )
      {}

      ~gzstreambuf ()
      {
        writeData(
#ifdef HAVE_LIBZ
                  Z_FINISH
#endif
                  );

#ifdef HAVE_LIBZ
        if( M_compression_stream )
          deflateEnd( M_compression_stream );
        if( M_decompression_stream )
          deflateEnd( M_decompression_stream );
#endif

        delete M_output_stream;
        delete M_input_stream;
        delete[] M_read_buffer;
        delete[] M_input_buffer;
        delete[] M_output_buffer;
        delete[] M_write_buffer;
#ifdef HAVE_LIBZ
        delete M_compression_stream;
        delete M_decompression_stream;
#endif
      }

      bool
      setLevel( int level )
      { 
#ifdef HAVE_LIBZ
        if( level <= Z_BEST_COMPRESSION ) 
          {
            sync(); // make sure there is room for the deflate to flush
            if( level >= 0 )
              {
                if( M_compression_stream )
                  deflateParams( M_compression_stream, level, Z_DEFAULT_STRATEGY );
                sync(); // write any data that was flushed by deflateParams
              }
            M_level = level;
            return true;
          }
        else
          return false;
#else
        if( M_level >= 0 )
          return false;
        else
          return true;
#endif
      }

    protected:
      bool
      writeData(
                int sync
#ifdef HAVE_LIBZ 
                = Z_NO_FLUSH
#else
                = 0
#endif
                )
      {
        int size = (pptr () - M_output_buffer) * sizeof(char_type);
        if( size == 0 )
          return true;


        if( M_output_stream == NULL )
          M_output_stream = new std::ostream( &M_strmbuf );

#ifdef HAVE_LIBZ
        if( M_level < 0 )
#endif
          M_output_stream->write( M_output_buffer, size );
#ifdef HAVE_LIBZ
        else
          {
            if( M_compression_stream == NULL )
              {
                M_compression_stream = new z_stream;
                M_compression_stream->zalloc = Z_NULL;
                M_compression_stream->zfree = Z_NULL;
                M_compression_stream->opaque = NULL;
                M_compression_stream->avail_in = 0;
                M_compression_stream->next_in = 0;
                M_compression_stream->next_out = 0;
                M_compression_stream->avail_out = 0;
                if( deflateInit( M_compression_stream, M_level ) != Z_OK )
                  {
 //                     cerr << "error in init\n";
                    return false;
                  }
                if( M_write_buffer == NULL )
                  M_write_buffer = new char[ M_buffer_size ];
                M_compression_stream->next_out = (Bytef*)M_write_buffer;
                M_compression_stream->avail_out = M_buffer_size;
              }
            M_compression_stream->next_in = (Bytef*)M_output_buffer;
            M_compression_stream->avail_in = size;

            do
              {
                int bytes_out = -M_compression_stream->total_out;
                int err = deflate( M_compression_stream, sync );
                if( err != Z_OK && err != Z_STREAM_END )
                  {
//                      cerr << "error deflating\n";
//                      cerr << M_compression_stream->msg << endl;
                    return false;
                  }
                bytes_out += M_compression_stream->total_out;
                M_output_stream->write( M_write_buffer, bytes_out ); 
                M_compression_stream->next_out = (Bytef*)M_write_buffer;
                M_compression_stream->avail_out = M_buffer_size;
              }
            while( M_compression_stream->avail_in != 0 );
            // we want to keep writing until all the data has been
            // consumed by the compression stream
          }
#endif
        if( sync != 
#ifdef HAVE_LIBZ
            Z_NO_FLUSH 
#else
            0
#endif
            ) 
          M_output_stream->flush(); // flush the underlying stream if a flush has
        // been used
        return true;
      }

      int
      readData( char* dest, int& dest_size )
      {
          if( M_input_stream == NULL )
              M_input_stream = new std::istream( &M_strmbuf );
          
          int rval = 0;

          if( !M_input_stream->good() )
          {
              M_input_stream->setstate( std::istream::failbit );
              return rval;
          }

          int bytes_to_read = M_strmbuf.in_avail();

          if( bytes_to_read < 0 )
              M_input_stream->setstate( std::istream::eofbit );
          else if( bytes_to_read == 0 )
          {
              rval = M_input_stream->read( dest, 1 ).gcount();
              bytes_to_read = M_strmbuf.in_avail();
              if( bytes_to_read > dest_size - 1 )
                  bytes_to_read = dest_size - 1;
              rval += M_input_stream->read( dest + 1, bytes_to_read ).gcount();
              dest_size = bytes_to_read + 1;
          }
          else
          {
              if( bytes_to_read > dest_size )
                  bytes_to_read = dest_size;
              rval = M_input_stream->read( dest, bytes_to_read ).gcount();
              dest_size = bytes_to_read;
          }
          return rval;
      }

      int_type
      overflow( int_type c = EOF )
      {
        // this method is supposed to flush the put area of the buffer
        // to the I/O device

        // if the buffer was not already allocated nor set by user,
        // do it just now
        if( pptr() == NULL )
          {
            M_output_buffer = new char_type[ M_buffer_size ];
          }
        else
          {
            if( !writeData() )
              return EOF;
          }
        setp( M_output_buffer, M_output_buffer + M_buffer_size );
        if( c != EOF )
          sputc( c );
        return 0;
      }

      int
      sync()
      {
        if( pptr() != NULL )
          {
            // just flush the put area
            if( !writeData(
#ifdef HAVE_LIBZ
                           Z_SYNC_FLUSH
#else
                           1
#endif
                           ) )
              return -1;
            setp( M_output_buffer, M_output_buffer + M_buffer_size );
          }
        return 0;
      }

      int_type
      underflow()
      {
        // this method is supposed to read some bytes from the I/O device

        // if the buffer was not already allocated nor set by user,
        // do it just now
        if( gptr() == NULL )
          {
            M_input_buffer = new char_type[ M_buffer_size ];
          }

        int readn = 0;
#ifdef HAVE_LIBZ
        if( M_level < 0 )
          {
#endif
            if( M_remained != 0 )
              M_input_buffer[ 0 ] = M_remaining_char;

            readn = M_buffer_size * sizeof(char_type) - M_remained;
            readData( M_input_buffer + M_remained, readn );
            int totalbytes = readn + M_remained;
            setg( M_input_buffer, M_input_buffer,
                  M_input_buffer + totalbytes / sizeof( char_type ) );
        
            M_remained = totalbytes % sizeof( char_type );
            if( M_remained != 0 )
              M_remaining_char = M_input_buffer[ totalbytes / sizeof(char_type) ];
#ifdef HAVE_LIBZ
          }
        else
          {
            if( M_read_buffer == NULL )
              M_read_buffer = new char_type[ M_buffer_size ];
          
        
            if( M_remained != 0 )
              M_input_buffer[ 0 ] = M_remaining_char;
  
      
            if( M_decompression_stream == NULL )
              {
                M_decompression_stream = new z_stream;
                M_decompression_stream->zalloc = Z_NULL;
                M_decompression_stream->zfree = Z_NULL;
                M_decompression_stream->opaque = NULL;
                M_decompression_stream->avail_in = 0;
                M_decompression_stream->next_in = 0;
                M_decompression_stream->avail_out = 0;
                M_decompression_stream->next_out = 0;    
                if( inflateInit( M_decompression_stream ) != Z_OK )
                  return false;
              }
            M_decompression_stream->next_out = (Bytef*)( M_input_buffer
                                                         + M_remained );
            M_decompression_stream->avail_out = ( M_buffer_size 
                                                  * sizeof(char_type) 
                                                  - M_remained );
            do
              {
                if( M_decompression_stream->avail_in == 0 )
                  {
                    int bytes_read = M_buffer_size;
                    readData( M_read_buffer, bytes_read );
                
                    M_decompression_stream->next_in = (Bytef*)M_read_buffer;
                    M_decompression_stream->avail_in = bytes_read;
                  }
                       
                readn -= M_decompression_stream->total_out;
                
                if( inflate( M_decompression_stream, Z_NO_FLUSH ) != Z_OK )
                  {
 //                     cerr << M_decompression_stream->msg << endl;
                    return EOF;
                  }
                readn += M_decompression_stream->total_out;
            
              }
            while( readn == 0 );
            int totalbytes = readn + M_remained;
            setg( M_input_buffer, M_input_buffer,
                  M_input_buffer + totalbytes / sizeof( char_type ) );
        
            M_remained = totalbytes % sizeof( char_type );
            if( M_remained != 0 )
              M_remaining_char = M_input_buffer[ ( totalbytes 
                                                   / sizeof( char_type ) ) ];
          }
#endif
        // if (readn == 0 && remained_ != 0)
        // error - there is not enough bytes for completing
        // the last character before the end of the stream
        // - this can mean error on the remote end
        if( readn == 0 )
          return EOF;

        return sgetc();   
      }

    private:

      // not for use
      gzstreambuf(const gzstreambuf&);
      gzstreambuf& operator=(const gzstreambuf&);

      std::streambuf& M_strmbuf;
      std::ostream* M_output_stream;     // used for writing to M_strmbuf
      std::istream* M_input_stream;      // used for reading from M_strmbuf
      std::streamsize M_buffer_size; // size of the following buffers
      char_type *M_read_buffer;     // used to read compressed data from M_strmbuf
      char_type *M_input_buffer;    // used to buffer uncompressed input to
                                // this stream
      char_type *M_output_buffer;   // used to buffer uncompressed output from
                                // this stream 
      char_type *M_write_buffer;    // used to write compressed data to M_strmbuf

      int M_remained;               // number of bytes remaining in M_output_buffer
      char_type M_remaining_char;

#ifdef HAVE_LIBZ
      z_stream* M_compression_stream; // used to perform the compression
      z_stream* M_decompression_stream; // used to perform the decompression
#endif
      int M_level;                  // current level of compression/decompression
                                // a level of -1 indicates that data is read
                                // and written without modification.
    };

    class gzstream
      : public gzstreambuf,
        public std::iostream
    {
    public:

      gzstream( std::streambuf& strm, unsigned int buffer_size = 8192 )
        : gzstreambuf( strm, buffer_size ),
          std::iostream( this )
      {
      }

      gzstream( std::iostream& strm, unsigned int buffer_size = 8192 )
        : gzstreambuf( *(strm.rdbuf()), buffer_size ),
          std::iostream( this )
      {
      }

    private:
      // not for use
      gzstream( const gzstream& );
      gzstream& operator=( const gzstream& );
    };


    class gzistream
      : public gzstreambuf,
        public std::istream
    {
    public:

      gzistream( std::streambuf& src, unsigned int buffer_size = 8192 )
        : gzstreambuf( src, buffer_size ),
          std::istream( this )
      {
      }

      gzistream( std::istream& src, unsigned int buffer_size = 8192 )
        : gzstreambuf( *(src.rdbuf()), buffer_size ),
          std::istream( this )
      {
      }

    private:
      // not for use
      gzistream( const gzistream& );
      gzistream& operator=( const gzistream& );
    };

    class gzostream
      : public gzstreambuf,
        public std::ostream
    {
    public:

      gzostream( std::streambuf& dest, unsigned int buffer_size = 8192 )
        : gzstreambuf( dest, buffer_size ),
          std::ostream( this )
      {
      }

      gzostream( std::ostream& dest, unsigned int buffer_size = 8192 )
        : gzstreambuf( *(dest.rdbuf()), buffer_size ),
          std::ostream( this )
      {
      }

    private:
      // not for use
      gzostream( const gzostream& );
      gzostream& operator=( const gzostream& );
    };
  } // namespace gz
} // namespace rcss
#endif 
