// -*-c++-*-

/***************************************************************************
                                libloader.h 
                             -------------------
                             dlopen a library
    begin                : 2002-10-10
    copyright            : (C) 2002 by The RoboCup Soccer Simulator 
                           Maintenance Group.
    email                : sserver-admin@lists.sourceforge.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU LGPL as published by the Free Software  *
 *   Foundation; either version 2 of the License, or (at your option) any  *
 *   later version.                                                        *
 *                                                                         *
 ***************************************************************************/


#ifndef RCSSLIBLOADER_H
#define RCSSLIBLOADER_H

#include <map>
#include <stack>
#include <list>
#include <exception>
#include <string>
#include <boost/shared_ptr.hpp>
#include <boost/weak_ptr.hpp>
#include <iostream>


namespace rcss
{
    class LibNotFound
        : public std::exception
    {
    public:
        LibNotFound( const char* msg ) throw()
            : m_msg( msg )
        {}
        
        ~LibNotFound() throw()
        {}
        
        const char*
        what() const throw()
        { 
            return m_msg.c_str();
        }
        
    private:
        std::string m_msg;
    };
 
    class LibInitError
        : public std::exception
    {
    public:
        LibInitError() throw()
        {}
        
        ~LibInitError() throw()
        {}
        
        const char*
        what() const throw()
        { return "Could not initialize library\n"; }
    };


    class LibLoaderImpl
    {
    public:
        virtual
        ~LibLoaderImpl();

        std::string
        name() const
        { return m_name; }

        std::string
        strippedName() const
        { return m_stripped_name; }

    protected:
        LibLoaderImpl( const std::string& );
        
        void
        load();

        void
        close();
    private:
        virtual
        void
        doLoad() = 0;

        virtual
        void
        doClose() = 0;

        virtual
        void*
        doGetSymb( const std::string& ) const = 0;

        void
        initialize();

        void
        finalize();

        friend class LibLoader;

    private:
        std::string m_name;
        std::string m_stripped_name;
    };

    typedef boost::shared_ptr< LibLoaderImpl > Lib;
    typedef boost::weak_ptr< LibLoaderImpl > WeakLib;

    struct StaticLib
    {
        const char* name;
        bool(*init)();
        void(*fin)();
    };
  
    class LibLoader
    {
    public:
        class StaticImporter
        {
        public:
            StaticImporter( const StaticLib& static_lib )
            { LibLoader::addStaticLib( static_lib ); }
        };
    
        friend class StaticImporter;
    
    private:
        typedef std::map< std::string, WeakLib > CacheMap;
        typedef std::stack< std::pair< Lib, WeakLib > > DepMap;

    public:
        LibLoader();
        ~LibLoader();

        static
        Lib
        load( const std::string& lib );

        static
        Lib
        loadExt( const std::string& lib );

        static
        bool
        addPath( const std::string& path );

        static
        bool
        setPath( const std::string& path );

        static
        std::string
        getPath();

        static
        Lib
        loadFromCache( const std::string& lib );

        static
        std::string
        strip( const std::string& filename );

        static
        std::string
        stripDirName( const std::string& filename );
        
        static
        std::string
        stripExt( const std::string& filename );

    private:
        static
        void
        addToCache( const std::string& lib_name, const Lib& lib );

        static
        void
        addDep( const Lib& lib );

        static
        void
        addStaticLib( const StaticLib& static_lib );

        static
        LibLoader&
        instance();
     
        friend class LibLoaderImpl;
        friend class StaticLibLoader;

    private:
        std::list< const StaticLib* > m_static_libs;
        CacheMap m_cached_libs;
        DepMap m_deps;
    };
}


// Why a C interface?  Only to make it easy for autoconf to detect the
// library.
extern "C" {
    const char*
    rcss_LibLoader_version();
}

#if defined(_WIN32) || defined(__WIN32__) || defined WIN32 || defined (__CYGWIN__)
#  define LL_EXPORT __declspec(dllexport)
#else
#  define LL_EXPORT
#endif


#define LIBLOADER_EXPORT( modulename, initfunc, finfunc ) \
extern "C" \
{ \
    LL_EXPORT bool \
    modulename##_initialize() \
    { return initfunc(); } \
\
    LL_EXPORT void \
    modulename##_finalize() \
    { finfunc(); } \
}


#endif
